from typing import Sequence

import numpy as np
import pandas as pd
import pytest

from ultrack.tracks.stats import (
    estimate_drift,
    tracks_df_movement,
    tracks_length,
    tracks_profile_matrix,
)
from ultrack.utils.constants import NO_PARENT


def spatial_df(group_drift: Sequence[int], length_per_group: int = 10) -> pd.DataFrame:
    rng = np.random.default_rng(42)

    df = []
    for d in group_drift:
        pos = np.empty((length_per_group, 4))
        pos[:, 0] = np.arange(length_per_group)
        pos[0, 1:] = rng.uniform(size=3)

        for i in range(length_per_group - 1):
            step = rng.uniform(size=3)
            pos[i + 1, 1:] = pos[i, 1:] + d * step / np.linalg.norm(step)
        df.append(pd.DataFrame(pos, columns=["t", "z", "y", "x"]))
    df = pd.concat(df)
    return df


@pytest.mark.parametrize(
    "drift",
    [0.5, 5, 10],
)
def test_spatial_drift(drift: float) -> None:
    df = spatial_df([drift])
    estimated_drift = np.linalg.norm(tracks_df_movement(df, lag=1), axis=1)
    assert np.allclose(estimated_drift[1:], drift)
    assert np.allclose(estimated_drift[:1], 0)


@pytest.mark.parametrize(
    "length_per_group",
    [3, 8, 15],
)
def test_maximum_distance_estimate(length_per_group: int) -> None:
    group_drift = [1, 10, 20]
    df = spatial_df(group_drift, length_per_group)
    df["track_id"] = np.repeat(np.arange(len(group_drift)), length_per_group)
    assert np.allclose(estimate_drift(df), max(group_drift))


def test_tracks_df_movement():
    # Sample test data
    df = pd.DataFrame(
        {
            "track_id": [1, 1, 2, 2],
            "t": [1, 2, 1, 2],
            "z": [0, 1, 0, 2],
            "y": [1, 2, 1, 2],
            "x": [2, 3, 2, 2],
        }
    )

    # Call the function
    result = tracks_df_movement(df)

    # Expected result
    expected = pd.DataFrame(
        {
            "z": [0.0, 1.0, 0.0, 2.0],
            "y": [0.0, 1.0, 0.0, 1.0],
            "x": [0.0, 1.0, 0.0, 0.0],
        }
    )

    # Assert that the result matches the expected dataframe
    pd.testing.assert_frame_equal(result, expected)


def test_tracks_df_movement_2d():
    # Sample test data
    df = pd.DataFrame(
        {
            "track_id": [1, 1, 2, 2],
            "t": [1, 2, 1, 2],
            "y": [1, 2, 1, 2],
            "x": [2, 3, 2, 2],
        }
    )

    # Call the function
    result = tracks_df_movement(df)

    # Expected result
    expected = pd.DataFrame(
        {
            "y": [0.0, 1.0, 0.0, 1.0],
            "x": [0.0, 1.0, 0.0, 0.0],
        }
    )

    # Assert that the result matches the expected dataframe
    pd.testing.assert_frame_equal(result, expected)


def test_tracks_profile_matrix_one_track_one_attribute():
    tracks_df = pd.DataFrame(
        {"track_id": [1, 1, 1], "t": [0, 1, 2], "attribute_1": [10, 20, 30]}
    )
    columns = ["attribute_1"]
    result = tracks_profile_matrix(tracks_df, columns)
    expected_result = np.array([[10, 20, 30]])
    assert np.array_equal(result, expected_result)


def test_tracks_profile_matrix_multiple_tracks_multiple_attributes():
    tracks_df = pd.DataFrame(
        {
            "track_id": [1, 1, 2, 2, 2],
            "t": [0, 1, 0, 1, 2],
            "attribute_1": [10, 20, 30, 40, 50],
            "attribute_2": [100, 200, 300, 400, 500],
        }
    )
    columns = ["attribute_1", "attribute_2"]
    result = tracks_profile_matrix(tracks_df, columns)
    expected_result = np.array(
        [[[10, 20, 0], [30, 40, 50]], [[100, 200, 0], [300, 400, 500]]]
    )
    assert np.array_equal(result, expected_result)


def test_tracks_profile_matrix_missing_timesteps():
    tracks_df = pd.DataFrame(
        {
            "track_id": [1, 1, 2, 2, 2],
            "t": [0, 1, 0, 2, 3],
            "attribute_1": [10, 20, 30, 40, 50],
            "attribute_2": [100, 200, 300, 400, 500],
        }
    )
    columns = ["attribute_1", "attribute_2"]
    result = tracks_profile_matrix(tracks_df, columns)
    expected_result = np.array(
        [[[10, 20, 0, 0], [30, 0, 40, 50]], [[100, 200, 0, 0], [300, 0, 400, 500]]]
    )
    assert np.array_equal(result, expected_result)


def test_tracks_profile_matrix_missing_attributes():
    tracks_df = pd.DataFrame(
        {
            "track_id": [1, 1, 2, 2, 2],
            "t": [0, 1, 0, 1, 2],
            "attribute_1": [10, 20, 30, 40, 50],
        }
    )
    columns = ["attribute_1", "attribute_2"]
    with pytest.raises(ValueError):
        tracks_profile_matrix(tracks_df, columns)


def test_tracks_length():
    # Sample data
    data = {
        "t": [0, 1, 2, 3, 4, 3, 4, 5, 2, 3, 4, 5, 4, 5],
        "track_id": [1, 1, 1, 2, 2, 3, 3, 3, 4, 4, 5, 5, 6, 6],
        "parent_track_id": [
            NO_PARENT,
            NO_PARENT,
            NO_PARENT,
            1,
            1,
            1,
            1,
            1,
            NO_PARENT,
            NO_PARENT,
            4,
            4,
            4,
            4,
        ],
    }
    df = pd.DataFrame(data)

    length_df = tracks_length(df)
    np.testing.assert_array_equal(length_df["length"], [3, 2, 3, 2, 2, 2])
    np.testing.assert_array_equal(length_df["start"], [0, 3, 3, 2, 4, 4])
    np.testing.assert_array_equal(length_df["end"], [2, 4, 5, 3, 5, 5])

    length_df = tracks_length(df, include_appearing=False)
    assert 4 not in length_df["track_id"].to_numpy()
    np.testing.assert_array_equal(length_df["length"], [3, 2, 3, 2, 2])
    np.testing.assert_array_equal(length_df["start"], [0, 3, 3, 4, 4])
    np.testing.assert_array_equal(length_df["end"], [2, 4, 5, 5, 5])

    length_df = tracks_length(df, include_disappearing=False)
    assert 2 not in length_df["track_id"].to_numpy()
    np.testing.assert_array_equal(length_df["length"], [3, 3, 2, 2, 2])
    np.testing.assert_array_equal(length_df["start"], [0, 3, 2, 4, 4])
    np.testing.assert_array_equal(length_df["end"], [2, 5, 3, 5, 5])

    length_df = tracks_length(df, include_disappearing=False, include_appearing=False)
    assert 2 not in length_df["track_id"].to_numpy()
    assert 4 not in length_df["track_id"].to_numpy()
    np.testing.assert_array_equal(length_df["length"], [3, 3, 2, 2])
    np.testing.assert_array_equal(length_df["start"], [0, 3, 4, 4])
    np.testing.assert_array_equal(length_df["end"], [2, 5, 5, 5])



import toml
from magicgui.widgets import FileEdit, PushButton, create_widget
from napari.layers import Image

from ultrack.config import MainConfig
from ultrack.widgets.ultrackwidget._legacy.baseconfigwidget import BaseConfigWidget


class MainConfigWidget(BaseConfigWidget):
    def __init__(self, config: MainConfig):
        super().__init__(label="ULTRACK", config=config)

    def _setup_widgets(self) -> None:
        self._foreground_layer_w = create_widget(annotation=Image, label="Foreground")
        self.append(self._foreground_layer_w)

        self._edge_layer_w = create_widget(annotation=Image, label="Edge")
        self.append(self._edge_layer_w)

        self._config_loader_w = FileEdit(
            filter="*.toml", label="Config. path", value="config.toml"
        )
        self.append(self._config_loader_w)

        self._save_config_btn = PushButton(text="Save config")
        self._save_config_btn.changed.connect(self._on_save_config)
        self.append(self._save_config_btn)

    def _on_save_config(self) -> None:
        with open(self._config_loader_w.value, mode="w") as f:
            toml.dump(self.config.model_dump(by_alias=True), f)



import logging
from typing import Optional

import edt
import numpy as np
from numpy.typing import ArrayLike
from skimage.morphology import reconstruction

from ultrack.imgproc.utils import _channel_iterator, _parse_voxel_size
from ultrack.utils.constants import ULTRACK_DEBUG
from ultrack.utils.cuda import import_module, is_cupy_array, to_cpu, xp

LOG = logging.getLogger(__name__)


def reconstruction_by_dilation(
    seed: ArrayLike, mask: ArrayLike, iterations: int
) -> ArrayLike:
    """
    Morphological reconstruction by dilation.
    This function does not compute the full reconstruction.
    The reconstruction is garanteed to be computed fully if the
    number of iterations is equal to the number of pixels.
    See: scikit-image docs for details.

    Parameters
    ----------
    seed : ArrayLike
        Dilation seeds.
    mask : ArrayLike
        Guidance mask.
    iterations : int
        Number of dilations.

    Returns
    -------
        Image reconstructed by dilation.
    """
    ndi = import_module("scipy", "ndimage", arr=mask)

    seed = np.minimum(seed, mask, out=seed)  # just making sure

    for _ in range(iterations):
        seed = ndi.grey_dilation(seed, size=3, output=seed, mode="constant")
        seed = np.minimum(seed, mask, out=seed)

    return seed


def detect_foreground(
    image: ArrayLike,
    voxel_size: Optional[ArrayLike] = None,
    sigma: float = 15.0,
    remove_hist_mode: bool = False,
    min_foreground: float = 0.0,
    channel_axis: Optional[int] = None,
) -> np.ndarray:
    """
    Detect foreground using morphological reconstruction by dilation and thresholding.

    Parameters
    ----------
    image : ArrayLike
        Input image.
    voxel_size : ArrayLike
        Array of voxel size (z, y, x).
    sigma : float, optional
        Sigma used to estimate background, it will be divided by voxel size, by default 15.0
    remove_hist_mode : bool, optional
        Removes histogram mode before computing otsu threshold, useful when background regions are being detected.
    min_foreground : float, optional
        Minimum value of foreground pixels after background subtraction and smoothing, by default 0.0
    channel_axis : Optional[int], optional
        When provided it will be used to compute the foreground mask for each channel separately and merge them.

    Returns
    -------
    ArrayLike
        Binary foreground mask.
    """
    ndi = import_module("scipy", "ndimage")
    filters = import_module("skimage", "filters")
    exposure = import_module("skimage", "exposure")

    voxel_size = _parse_voxel_size(voxel_size, image.ndim, channel_axis)

    shape = list(image.shape)
    if channel_axis is not None:
        shape.pop(channel_axis)

    sigmas = sigma / voxel_size
    LOG.info(f"Detecting foreground with voxel size {voxel_size} and sigma {sigma}")
    LOG.info(f"Sigmas after scaling {sigmas}")

    output = np.zeros(shape, dtype=bool)

    for ch in _channel_iterator(image, channel_axis):
        if ch is None:
            _image = image
        else:
            _image = np.take(image, indices=ch, axis=channel_axis)

        _image = xp.asarray(_image)

        seed = ndi.gaussian_filter(_image, sigma=sigmas)
        if is_cupy_array(_image):
            background = reconstruction_by_dilation(seed, _image, 100)
        else:
            if seed.ndim > 2:
                LOG.warning(
                    "Using CPU background reconstruction, this could take a while, consider using GPU for 3D images."
                )
            np.minimum(seed, _image, out=seed)  # required condition for reconstruction
            background = reconstruction(seed, _image, method="dilation")
        del seed

        foreground = _image - background
        del background

        # threshold in smaller image to save memory and sqrt to deskew data distribution towards left
        small_foreground = np.sqrt(
            ndi.zoom(foreground, (0.25,) * foreground.ndim, order=1)
        )

        # begin thresholding
        robust_max = np.quantile(small_foreground, 1 - 1e-6)
        small_foreground = np.minimum(
            robust_max, small_foreground, out=small_foreground
        )

        # number of bins according to maximum value
        nbins = int(robust_max / 10)  # binning with window of 10
        nbins = min(nbins, 256)
        nbins = max(nbins, 10)

        LOG.info(f"Estimated almost max. {np.square(robust_max)}")
        LOG.info(f"Histogram with {nbins}")

        hist, bin_centers = exposure.histogram(small_foreground, nbins)

        # histogram disconsidering pixels we are sure are background
        if remove_hist_mode:
            remaining_background_idx = hist.argmax() + 1
            hist = hist[remaining_background_idx:]
            bin_centers = bin_centers[remaining_background_idx:]

        del small_foreground
        threshold = np.square(filters.threshold_otsu(hist=(hist, bin_centers)))
        LOG.info(f"Threshold {threshold}")

        threshold = max(threshold, min_foreground)
        LOG.info(f"Threshold after minimum filtering {threshold}")

        mask = foreground > threshold
        del foreground

        struct = ndi.generate_binary_structure(mask.ndim, 2)
        mask = ndi.binary_opening(mask, structure=struct, output=mask)
        # opening but with border value=True on erosion
        mask = ndi.binary_dilation(mask, structure=struct, output=mask)
        mask = ndi.binary_erosion(
            mask, structure=struct, output=mask, border_value=True
        )
        output |= to_cpu(mask)

    if ULTRACK_DEBUG:
        import napari

        viewer = napari.Viewer()
        viewer.add_image(to_cpu(_image))
        viewer.add_labels(output)
        napari.run()

    return output


def inverted_edt(
    mask: ArrayLike,
    voxel_size: Optional[ArrayLike] = None,
    axis: Optional[int] = None,
) -> ArrayLike:
    """
    Computes Euclidean distance transform (EDT), inverts and normalizes it.

    Parameters
    ----------
    mask : ArrayLike
        Cells' foreground mask.
    voxel_size : Optional[ArrayLike], optional
        Voxel size, by default None
    axis : Optional[int], optional
        Axis to compute the EDT, by default None

    Returns
    -------
    ArrayLike
        Inverted and normalized EDT.
    """
    mask = np.asarray(mask)
    if axis is None:
        dist = edt.edt(mask, anisotropy=voxel_size)
    else:
        dist = np.stack(
            [
                edt.edt(np.take(mask, i, axis=axis), anisotropy=voxel_size)
                for i in range(mask.shape[axis])
            ],
            axis=axis,
        )
    dist = dist / dist.max()
    dist = 1.0 - dist
    dist[mask == 0] = 1
    return dist


class Cellpose:
    def __init__(self, **kwargs) -> None:
        """See cellpose.models.Cellpose documentation for details."""
        from cellpose.models import CellposeModel as _Cellpose

        if "pretrained_model" not in kwargs and "model_type" not in kwargs:
            kwargs["model_type"] = "cyto"

        self.model = _Cellpose(**kwargs)

    def __call__(self, image: ArrayLike, **kwargs) -> np.ndarray:
        """
        Predicts image labels.
        See cellpose.models.Cellpose.eval documentation for details.
        """
        labels, _, _ = self.model.eval(image, **kwargs)
        return labels



import logging
from functools import wraps
from typing import Optional, Sequence, Tuple, Union

from numpy.typing import ArrayLike
from tqdm import tqdm
from zarr.storage import StoreLike

from ultrack.utils.array import create_zarr
from ultrack.utils.cuda import import_module, to_cpu, xp
from ultrack.utils.deprecation import rename_argument

LOG = logging.getLogger(__name__)


@rename_argument("detection_store_or_path", "foreground_store_or_path")
@rename_argument("edges_store_or_path", "contours_store_or_path")
def labels_to_contours(
    labels: Union[ArrayLike, Sequence[ArrayLike]],
    sigma: Optional[Union[Sequence[float], float]] = None,
    foreground_store_or_path: Optional[StoreLike] = None,
    contours_store_or_path: Optional[StoreLike] = None,
    overwrite: bool = False,
) -> Tuple[ArrayLike, ArrayLike]:
    """
    Converts and merges a sequence of labels into ultrack input format (foreground and contours)

    Parameters
    ----------
    labels : Union[ArrayLike, Sequence[ArrayLike]]
        List of labels with equal shape.
    sigma : Optional[Union[Sequence[float], float]], optional
        Contours smoothing parameter (gaussian blur), contours aren't smoothed when not provided.
    foreground_store_or_path : str, zarr.storage.Store, optional
        Zarr storage, it can be used with zarr.storage.LocalStore to save the output into disk.
        By default it creates a temporary store.
    contours_store_or_path : str, zarr.storage.Store, optional
        Zarr storage, it can be used with zarr.storage.LocalStore to save the output into disk.
        By default it creates a temporary store.
    overwrite : bool, optional
        Overwrite output output files if they already exist, by default False.

    Returns
    -------
    Tuple[ArrayLike, ArrayLike]
        Combined foreground and edges arrays.
    """
    ndi = import_module("scipy", "ndimage")
    segm = import_module("skimage", "segmentation")

    if not isinstance(labels, Sequence):
        labels = [labels]

    shape = labels[0].shape
    for lb in labels:
        if shape != lb.shape:
            raise ValueError(
                f"All labels must have the same shape. Found {shape} and {lb.shape}"
            )

    LOG.info(f"Labels shape {shape}")

    foreground = create_zarr(
        shape=shape,
        dtype=bool,
        store_or_path=foreground_store_or_path,
        overwrite=overwrite,
        default_store_type=None,
    )
    contours = create_zarr(
        shape=shape,
        dtype=xp.float32,
        store_or_path=contours_store_or_path,
        overwrite=overwrite,
        default_store_type=None,
    )

    for t in tqdm(range(shape[0]), "Converting labels to contours"):
        foreground_frame = xp.zeros(shape[1:], dtype=foreground.dtype)
        contours_frames = xp.zeros(shape[1:], dtype=contours.dtype)

        for lb in labels:
            lb_frame = xp.asarray(lb[t])
            foreground_frame |= lb_frame > 0
            contours_frames += segm.find_boundaries(lb_frame, mode="outer")

        contours_frames /= len(labels)

        if sigma is not None:
            contours_frames = ndi.gaussian_filter(contours_frames, sigma)
            contours_frames = contours_frames / contours_frames.max()

        foreground[t] = to_cpu(foreground_frame)
        contours[t] = to_cpu(contours_frames)

    return foreground, contours


@wraps(labels_to_contours)
def labels_to_edges(*args, **kwargs) -> Tuple[ArrayLike, ArrayLike]:
    return labels_to_contours(*args, **kwargs)



from ultrack.widgets.division_annotation_widget import DivisionAnnotationWidget
from ultrack.widgets.hierarchy_viz_widget import HierarchyVizWidget
from ultrack.widgets.hypotheses_viz_widget import HypothesesVizWidget
from ultrack.widgets.labels_to_edges_widget import LabelsToContoursWidget
from ultrack.widgets.node_annotation_widget import NodeAnnotationWidget
from ultrack.widgets.track_inspection_widget import TrackInspectionWidget
from ultrack.widgets.ultrackwidget import UltrackWidget



import numpy as np
import pandas as pd
import zarr
import zarr.storage

from ultrack.tracks import close_tracks_gaps
from ultrack.utils.constants import NO_PARENT


def test_gap_closing() -> None:
    """

    track id's shown by numbers

    8      ---

    2   --  -7 -9
       /
    1-       - 5
       \\    /
    3   -- -   4
            \
             - 6
    """

    segments_with_gap = np.zeros((8, 7), dtype=np.uint8)

    # Create a DataFrame with tracks
    tracks_df_with_gap = pd.DataFrame(
        {
            "track_id": [1, 2, 2, 3, 3, 4, 5, 6, 7, 8, 8, 8, 9],
            "parent_track_id": [
                NO_PARENT,
                1,
                1,
                1,
                1,
                NO_PARENT,
                4,
                4,
                NO_PARENT,
                NO_PARENT,
                NO_PARENT,
                NO_PARENT,
                NO_PARENT,
            ],
            "t": [0, 1, 2, 1, 2, 4, 5, 5, 5, 4, 5, 6, 7.0],
            "x": [2, 0, 0, 4, 4, 4, 5, 6, 1, 3, 3, 3, 2.0],
        }
    )

    segments_with_gap[
        tracks_df_with_gap["t"].astype(int), tracks_df_with_gap["x"].astype(int)
    ] = tracks_df_with_gap["track_id"].values
    segments_with_gap = zarr.array(segments_with_gap)

    tracks_df, segments = close_tracks_gaps(
        tracks_df_with_gap,
        max_gap=2,
        max_radius=1.5,
        spatial_columns=["x"],
        segments=segments_with_gap,
    )
    expected_tracks_df = pd.DataFrame(
        {
            "track_id": [1, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 5, 6, 8, 8, 8],
            "parent_track_id": [
                NO_PARENT,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                3,
                3,
                NO_PARENT,
                NO_PARENT,
                NO_PARENT,
            ],
            "t": [0, 1, 2, 3, 4, 5, 6, 7, 1, 2, 3, 4, 5, 5, 4, 5, 6.0],
            "x": [2, 0, 0, 1 / 3, 2 / 3, 1, 1.5, 2, 4, 4, 4, 4, 5, 6, 3, 3, 3],
        }
    )

    expected_segments = np.zeros((8, 7), dtype=np.uint8)
    expected_segments[
        expected_tracks_df["t"].astype(int),
        np.floor(expected_tracks_df["x"]).astype(int),
    ] = expected_tracks_df["track_id"].values

    # +1 of 9 to 7
    # +2 of 7 to 2
    # +1 of 4 to 3
    assert tracks_df.shape[0] == tracks_df_with_gap.shape[0] + 1 + 2 + 1

    # forcing equal index
    expected_tracks_df.index = tracks_df.index

    pd.testing.assert_frame_equal(tracks_df, expected_tracks_df, check_like=True)

    np.testing.assert_array_equal(segments, expected_segments)



from typing import Optional

import click

from ultrack import MainConfig
from ultrack.api import start_server
from ultrack.cli.utils import config_option


@click.command("server")
@click.option("--host", help="Host address", default="0.0.0.0", show_default=True)
@click.option(
    "--port",
    help="Port number to listen on",
    type=int,
    default=61234,
    show_default=True,
)
@click.option(
    "--api-results-path",
    help="Path to the API results folder",
    type=str,
    default=None,
    show_default=True,
)
@config_option()
def server_cli(
    host: str, port: int, api_results_path: Optional[str], config: MainConfig
) -> None:
    """Start the websockets ultrack API."""
    start_server(api_results_path, ultrack_data_config=config, host=host, port=port)



import numpy as np
import pandas as pd
import pytest
import sqlalchemy as sqla
from sqlalchemy.orm import Session

from ultrack import solve, to_tracks_layer
from ultrack.config.config import MainConfig
from ultrack.core.database import LinkDB, NodeDB, VarAnnotation
from ultrack.core.solve.sqltracking import SQLTracking
from ultrack.utils.constants import NO_PARENT

_CONFIG_PARAMS = {
    "segmentation.n_workers": 4,
    "linking.n_workers": 4,
    "tracking.appear_weight": -0.25,
    "tracking.disappear_weight": -0.5,
    "tracking.division_weight": -0.25,
    "tracking.window_size": 5,
    "tracking.overlap_size": 2,
}

_TEST_PARAMS = [
    ({"data.database": "sqlite", **_CONFIG_PARAMS}, {"length": 10}),
    ({"data.database": "postgresql", **_CONFIG_PARAMS}, {"length": 10}),
]


def _validate_tracking_solution(config: MainConfig):
    nodes = pd.read_sql_query(
        f"SELECT id, t, selected, parent_id FROM {NodeDB.__tablename__}",
        con=config.data_config.database_path,
        index_col="id",
    )

    # at least one node was selected
    assert np.any(nodes["selected"])

    # assert parents id are valid
    assert np.all(
        nodes.loc[nodes["parent_id"] != NO_PARENT, "parent_id"].isin(nodes.index)
    )

    # assert starting nodes are parentless
    assert np.all(nodes.loc[nodes["t"] == 0, "parent_id"] == NO_PARENT)

    # assert there isn't any disconnected chunk
    for t, group in nodes.groupby("t"):
        if t == 0:
            continue
        assert np.any(group["parent_id"] != NO_PARENT)


@pytest.mark.parametrize(
    "config_content,timelapse_mock_data",
    _TEST_PARAMS,
    indirect=True,
)
def test_sql_tracking(
    linked_database_mock_data: MainConfig,
) -> None:
    config = linked_database_mock_data
    solve(config)
    _validate_tracking_solution(config)


@pytest.mark.parametrize(
    "config_content,timelapse_mock_data",
    _TEST_PARAMS,
    indirect=True,
)
def test_batch_sql_tracking(
    linked_database_mock_data: MainConfig,
) -> None:
    config = linked_database_mock_data

    solve(config, batch_index=0)
    solve(config, batch_index=1)

    with pytest.raises(ValueError):
        solve(config, batch_index=2)

    _validate_tracking_solution(config)


@pytest.mark.parametrize(
    "config_content",
    [
        {
            "data.database": "sqlite",
            "segmentation.n_workers": 4,
            "linking.n_workers": 4,
        },
        {
            "data.database": "postgresql",
            "segmentation.n_workers": 4,
            "linking.n_workers": 4,
        },
    ],
    indirect=True,
)
def test_clear_solution(
    tracked_database_mock_data: MainConfig,
) -> None:
    database_path = tracked_database_mock_data.data_config.database_path

    SQLTracking.clear_solution_from_database(database_path)

    engine = sqla.create_engine(database_path)
    with Session(engine) as session:
        assert session.query(LinkDB).count() > 0
        assert session.query(NodeDB).count() > 0
        assert session.query(NodeDB).where(NodeDB.selected).count() == 0
        assert session.query(NodeDB).where(NodeDB.parent_id != NO_PARENT).count() == 0


@pytest.mark.parametrize(
    "config_content,timelapse_mock_data",
    _TEST_PARAMS,
    indirect=True,
)
def test_annotations_sql_tracking(
    linked_database_mock_data: MainConfig,
) -> None:
    config = linked_database_mock_data

    solve(config, overwrite=True, use_annotations=True)
    tracks_df, _ = to_tracks_layer(config)

    engine = sqla.create_engine(config.data_config.database_path)
    with Session(engine) as session:
        session.query(NodeDB).where(NodeDB.t == 0).update(
            {"node_annot": VarAnnotation.FAKE}
        )
        session.commit()

    solve(config, overwrite=True, use_annotations=True)
    tracks_df_annot, _ = to_tracks_layer(config)

    assert len(tracks_df) > len(tracks_df_annot)



import logging
import shutil
from pathlib import Path
from typing import Any, Dict, Literal, Optional, Tuple

import numpy as np
import scipy.ndimage as ndi
import skimage.morphology as morph
from skimage.data import binary_blobs
from skimage.segmentation import find_boundaries, relabel_sequential, watershed

LOG = logging.getLogger(__name__)


def make_segmentation_mock_data(
    size: int = 64,
    n_dim: int = 3,
    rng: Optional[np.random.Generator] = None,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Generates mock segmentation dataset producing binary blobs, their contours and labels."""
    if rng is None:
        rng = np.random.default_rng(42)

    blobs = binary_blobs(length=size, n_dim=n_dim, volume_fraction=0.5, rng=rng)

    edt = ndi.distance_transform_edt(blobs)
    markers, _ = ndi.label(morph.h_maxima(edt, 2))
    labels = watershed(-edt, markers, mask=blobs)
    contours = find_boundaries(labels)
    labels, _, _ = relabel_sequential(labels)

    return blobs, contours, labels


def make_config_content(kwargs: Dict[str, Any] = {}) -> Dict[str, Any]:
    """Generates a mock configuration content dictionary."""
    content = {
        "data": {},
        "segmentation": {
            "threshold": 0.5,
            "max_area": 7500,
            "min_area": 500,
            "min_frontier": 0.1,
            "anisotropy_penalization": 0.0,
            "ws_hierarchy": "area",
            "n_workers": 1,
        },
        "linking": {
            "max_neighbors": 10,
            "max_distance": 15.0,
            "n_workers": 1,
        },
        "tracking": {
            "appear_weight": -0.2,
            "disappear_weight": -1.0,
            "division_weight": -0.1,
            "dismiss_weight_guess": None,
            "include_weight_guess": None,
            "solution_gap": 0.001,
            "time_limit": 36000,
            "method": -1,
            "n_threads": 0,
            "link_function": "identity",
            "solver_name": "GUROBI",
        },
    }

    for keys, value in kwargs.items():
        param = content
        keys = keys.split(".")
        for k in keys[:-1]:
            param = param[k]
        param[keys[-1]] = value

    LOG.info(content)

    return content


def make_cell_division_mock_data() -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Short timelapse of a sphere "dividing" into two.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray, np.ndarray]
        Foreground, contours, labels maps.
    """
    cells = np.zeros((5, 64, 64, 64), dtype=bool)
    ball = morph.ball(radius=5, dtype=bool)

    cells[(0,) + tuple(slice(27, 27 + s) for s in ball.shape)] |= ball

    cells[(1,) + tuple(slice(24, 24 + s) for s in ball.shape)] |= ball  # \
    cells[(1,) + tuple(slice(30, 30 + s) for s in ball.shape)] |= ball  # --- division

    cells[(2,) + tuple(slice(22, 22 + s) for s in ball.shape)] |= ball
    cells[(2,) + tuple(slice(32, 32 + s) for s in ball.shape)] |= ball

    cells[(3,) + tuple(slice(20, 20 + s) for s in ball.shape)] |= ball
    cells[(3,) + tuple(slice(35, 35 + s) for s in ball.shape)] |= ball  # \
    cells[(3,) + tuple(slice(29, 29 + s) for s in ball.shape)] |= ball  # --- division

    cells[(4,) + tuple(slice(18, 18 + s) for s in ball.shape)] |= ball
    cells[(4,) + tuple(slice(37, 37 + s) for s in ball.shape)] |= ball
    cells[(4,) + tuple(slice(27, 27 + s) for s in ball.shape)] |= ball

    contours = np.zeros_like(cells, dtype=np.float32)
    labels = np.zeros_like(cells, dtype=np.int32)

    for t in range(cells.shape[0]):
        edt = ndi.distance_transform_edt(cells[t])
        markers, _ = ndi.label(morph.h_maxima(edt, 2))
        label = watershed(-edt, markers, mask=cells[t])
        contours[t] = find_boundaries(label)
        labels[t] = label

    return cells, contours, labels


def validate_and_overwrite_path(
    path: Path, overwrite: bool, msg_type: Literal["cli", "api"]
) -> None:
    """Validates and errors existance of path (or dir) and overwrites it if requested."""

    if msg_type == "cli":
        msg = f"{path} already exists. Set `--overwrite` option to overwrite it."

    elif msg_type == "api":
        msg = f"{path} already exists. Set `overwrite=True` to overwrite it."

    else:
        raise ValueError(f"Invalid `msg_type` {msg_type}, must be `cli` or `api`.")

    if path.exists():
        if overwrite:
            if path.is_dir():
                shutil.rmtree(path)
            else:
                path.unlink()
        else:
            raise ValueError(msg)



import multiprocessing as mp
import uuid
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Callable, List, Optional, Sequence

import cloudpickle
import fasteners
from tqdm import tqdm

from ultrack.config.config import DataConfig
from ultrack.config.dataconfig import DatabaseChoices

"""
Configures multiprocessing to use multiprocessing for pickling.
This allows function to be pickled.
Reference: https://stackoverflow.com/a/69253561/6748803
"""
cloudpickle.Pickler.dumps, cloudpickle.Pickler.loads = (
    cloudpickle.dumps,
    cloudpickle.loads,
)
mp.connection._ForkingPickler = cloudpickle.Pickler


def multiprocessing_apply(
    func: Callable[[Any], None],
    sequence: Sequence[Any],
    n_workers: int,
    desc: Optional[str] = None,
) -> List[Any]:
    """Applies `func` for each item in `sequence`.

    Parameters
    ----------
    func : Callable[[Any], NoneType]
        Function to be executed.
    sequence : Sequence[Any]
        Sequence of parameters.
    n_workers : int
        Number of workers for multiprocessing.
    desc : Optional[str], optional
        Description to tqdm progress bar, by default None

    Returns
    -------
    List[int]
        List of `func` outputs.
    """
    length = len(sequence)
    if n_workers > 1 and length > 1:
        ctx = mp.get_context("spawn")
        with ctx.Pool(min(n_workers, length)) as pool:
            return list(tqdm(pool.imap(func, sequence), desc=desc, total=length))
    else:
        return [func(t) for t in tqdm(sequence, desc=desc)]


@contextmanager
def multiprocessing_sqlite_lock(
    data_config: DataConfig,
    n_workers: Optional[int] = None,
) -> Optional[fasteners.InterProcessLock]:
    """Write lock for writing on `sqlite` with multiprocessing. No lock otherwise."""

    lock = None
    is_parallel = n_workers is None or n_workers > 1  # unknown or > 1
    if data_config.database == DatabaseChoices.sqlite.value and is_parallel:
        identifier = uuid.uuid4().hex
        lock = fasteners.InterProcessLock(
            path=data_config.working_dir / f"{identifier}.lock"
        )

    try:
        yield lock

    finally:
        if lock is not None:
            Path(lock.path.decode("ascii")).unlink(missing_ok=True)


def batch_index_range(
    total_size: int, n_workers: int, batch_index: Optional[int]
) -> Sequence[int]:
    """Compute the range of the given `batch_index`, if None the whole `total_size` range is returned.

    Parameters
    ----------
    total_size : int
        Total size of the processing range.
    n_workers : int
        Number of workers per batch, processing window size.
    batch_index : Optional[int]
        Batch processing window index.

    Returns
    -------
    Sequence[int]
        Processing range of the given index.
    """
    if batch_index is None:
        return range(total_size)

    start = list(range(0, total_size, n_workers))[batch_index]
    return range(start, min(start + n_workers, total_size))



import os
import warnings

if os.environ.get("ULTRACK_DEBUG", False):
    import logging

    logger = logging.getLogger()
    logger.setLevel(logging.INFO)

# ignoring small float32/64 zero flush warning
warnings.filterwarnings("ignore", message="The value of the smallest subnormal for")

__version__ = "0.7.1"

from ultrack.config.config import MainConfig, load_config
from ultrack.core.export.ctc import to_ctc
from ultrack.core.export.exporter import export_tracks_by_extension
from ultrack.core.export.trackmate import to_trackmate
from ultrack.core.export.tracks_layer import to_tracks_layer
from ultrack.core.export.zarr import tracks_to_zarr
from ultrack.core.interactive import add_new_node
from ultrack.core.linking.processing import link
from ultrack.core.main import track
from ultrack.core.segmentation.processing import segment
from ultrack.core.solve.processing import solve
from ultrack.core.tracker import Tracker
from ultrack.imgproc.flow import add_flow



from enum import IntEnum

import numba
import numpy as np
from numba.experimental import jitclass


class Color(IntEnum):
    white = 0
    gray = 1
    black = 2


class Policy(IntEnum):
    minimum = 0
    maximum = 1


_SPEC = [
    ("values", numba.float32[:]),
    ("policy", numba.int8),
    ("_values", numba.float32[:]),
    ("_size", numba.int64),
    ("_age", numba.int64),
    ("_last", numba.int64),
    ("_policy", numba.int8),
    ("_nodes", numba.int64[:]),
    ("_pos", numba.int64[:]),
    ("_colors", numba.int8[:]),
    ("_ages", numba.int64[:]),
]


@jitclass(_SPEC)
class Heap:
    def __init__(self, values: np.ndarray, policy: Policy = Policy.maximum) -> None:
        """Argument heap, copied and modified from libift:
            https://github.com/PyIFT/pyift

        Parameters
        ----------
        values : np.ndarray
            Array of values used to heap sort, must be np.float32 (used by reference)
            for on-the-fly updates on the values.
        policy : Policy, optional
            Minimum or maximum policy, by default maximum.
        """
        size = len(values)
        self._values = values
        self._size = size
        self._age = 0
        self._last = -1
        self._policy = policy
        self._nodes = np.full(size, fill_value=-1, dtype=np.int64)
        self._pos = np.full(size, fill_value=-1, dtype=np.int64)
        self._colors = np.full(size, fill_value=Color.white, dtype=np.uint8)
        self._ages = np.zeros(size, dtype=np.int64)

    @staticmethod
    def _dad(i: int) -> int:
        return (i - 1) // 2

    @staticmethod
    def _left_son(i: int) -> int:
        return 2 * i + 1

    @staticmethod
    def _right_son(i: int) -> int:
        return 2 * i + 2

    def _swap(self, i: int, j: int) -> None:
        self._nodes[i], self._nodes[j] = self._nodes[j], self._nodes[i]
        self._pos[self._nodes[i]] = i
        self._pos[self._nodes[j]] = j

    def _lower(self, i: int, j: int) -> bool:
        ni, nj = self._nodes[i], self._nodes[j]
        if self._values[ni] == self._values[nj]:
            return self._ages[ni] < self._ages[nj]
        return self._values[ni] < self._values[nj]

    def _greater(self, i: int, j: int) -> bool:
        ni, nj = self._nodes[i], self._nodes[j]
        if self._values[ni] == self._values[nj]:
            return self._ages[ni] < self._ages[nj]  # always prioritize younger node
        return self._values[ni] > self._values[nj]

    def _go_up_heap_pos(self, pos: int) -> None:
        dad = self._dad(pos)
        if self._policy == Policy.minimum:
            while dad >= 0 and self._greater(dad, pos):
                self._swap(dad, pos)
                pos = dad
                dad = self._dad(pos)
        else:
            while dad >= 0 and self._lower(dad, pos):
                self._swap(dad, pos)
                pos = dad
                dad = self._dad(pos)

    def _go_down_heap_pos(self, pos: int) -> None:
        # numba wasn't compiling this function as a recursion
        # TODO: check if there's a better way to do this loop avoiding while True
        while True:
            next, left, right = pos, self._left_son(pos), self._right_son(pos)
            if self._policy == Policy.minimum:
                if left <= self._last and self._lower(left, next):
                    next = left
                if right <= self._last and self._lower(right, next):
                    next = right
            else:
                if left <= self._last and self._greater(left, next):
                    next = left
                if right <= self._last and self._greater(right, next):
                    next = right

            if next != pos:
                self._swap(next, pos)
                pos = next
            else:
                break

    def is_full(self) -> bool:
        return self._last == self._size - 1

    def is_empty(self) -> bool:
        return self._last == -1

    def insert(self, index: int) -> None:
        if self.is_full():
            raise ValueError("Heap is full.")

        # add node at last position
        self._last += 1
        self._age += 1
        self._nodes[self._last] = index
        self._ages[index] = self._age
        self._colors[index] = Color.gray
        self._pos[index] = self._last

        # sort heap
        self._go_up_heap_pos(self._last)

    def insert_array(self, array: np.ndarray) -> None:
        for i in array:
            self.insert(i)

    def pop(self) -> int:
        if self.is_empty():
            raise ValueError("Heap is empty.")

        # remove node at 0
        index = self._nodes[0]
        self._pos[index] = -1
        self._colors[index] = Color.black

        # swaps removed with last
        self._nodes[0] = self._nodes[self._last]
        self._pos[self._nodes[0]] = 0
        self._nodes[self._last] = -1
        self._last -= 1

        # sort heap
        self._go_down_heap_pos(0)

        return index

    def reset(self) -> None:
        self._nodes[...] = -1
        self._pos[...] = -1
        self._colors[...] = Color.white
        self._ages[...] = 0
        self._last = -1
        self._age = 0

    def go_up_heap(self, index: int) -> None:
        self._age += 1
        self._ages[index] = self._age
        self._go_up_heap_pos(self._pos[index])

    def go_down_heap(self, index: int) -> None:
        self._age += 1
        self._ages[index] = self._age
        self._go_down_heap_pos(self._pos[index])
